@@ -19,6 +19,8 @@ import pkg_resources
 import simplejson as json
 import yaml
 
+ureg = pint.UnitRegistry()
+
 
 def yamlcat(*files):
     """concat YAML files and returns a StringIO object
@@ -110,6 +112,53 @@ def load_yaml(*files, clean=True, debug=False):
     return specs
 
 
+def quantify_df(df, target_units, errors):
+    """preprocess a dataframe assuming units strings are on first line"""
+    df = df.copy()
+    df.columns = pd.MultiIndex.from_tuples(zip(df.columns, df.iloc[0]))
+    df = df.iloc[1:]  # drop first line which has been merged previously
+    _source_units = {}  # {col: u for col, u in df.columns if isinstance(u, str)}
+    _units_to_col = defaultdict(list)
+    # -------------------------------------------------------------------------
+    # split df between numeric / non-numeric columns
+    df_num = pd.DataFrame(index=df.index)
+    df_nonnum = pd.DataFrame(index=df.index)
+    # -------------------------------------------------------------------------
+    # test each column and column units
+    for col, col_units in df.columns:
+        if isinstance(col_units, str):
+            # found a column with a specified units.
+            # test specified units for being known by pint
+            if col_units.strip() != "":
+                try:
+                    getattr(ureg, col_units)
+                except AttributeError:
+                    errors[(col, None)].append("undefined units '%s'" % col_units)
+                    df_nonnum[(col, col_units)] = df[(col, col_units)]
+                    continue
+            # Try to convert the whole column # to numeric:
+            try:
+                df_num[(col, col_units)] = pd.to_numeric(df[(col, col_units)])
+            except ValueError as exc:
+                errors[(col, "?")] = "cannot convert some values to numeric"
+                df_nonnum[(col, col_units)] = df[(col, col_units)]
+                continue
+        else:
+            df_nonnum[(col, col_units)] = df[(col, col_units)]
+            if col in target_units:
+                errors[(col, None)].append("no units specified in source file")
+    # -------------------------------------------------------------------------
+    # calculate source units
+    for col, u in df_num.columns:
+        _source_units[col] = u
+        _units_to_col[u].append(col)
+    if len(df_num.columns) > 0:
+        df_num.columns = pd.MultiIndex.from_tuples(df_num.columns)
+    if len(df_nonnum.columns) > 0:
+        df_nonnum.columns = pd.MultiIndex.from_tuples(df_nonnum.columns)
+    return df_num, df_nonnum, _source_units, _units_to_col
+
+
 # use short units
 pintpandas.PintType.ureg.default_format = "~P"
 
@@ -321,59 +370,69 @@ class DataFrameSchematizer:
         self.columns_specs[name] = ref
 
     def validate_dataframe(self, df):
-        """validate dataframe against self.schema()"""
+        """validate dataframe against self.schema():
+
+        1. type validation
+        2. units conversion
+        3. default values are applied
+        4. final validation
+        """
         schema = self.build()
-        intermediate_report = defaultdict(list)  # what is not raised by jsonschema
+        validator = jsonschema.Draft7Validator(schema)
         # =====================================================================
-        # builds mult-header from dataframes if schemas are units-aware
+        # 1. Early validation for typing
+        # This is usefull since validation should occures at last,
+        # after units conversion and holes filling.
+        # If a data has the wrong type, this process would fail before reaching
+        # validation
+        # =====================================================================
+        initial_without_urow = df.copy()
+        if self._is_units:
+            initial_without_urow = initial_without_urow.drop(0)
+        # check **types only** before any further DataFrame transformation
+        # this will catch any type error
+        early_report = _validate(
+            json.loads(
+                json.dumps(initial_without_urow.to_dict(orient="list"), ignore_nan=True)
+            ),
+            validator,
+            check_types=True,
+        )
+        # =====================================================================
+        # builds multi-header from dataframes if schemas are units-aware
         # =====================================================================
         if self._is_units:
-            # -----------------------------------------------------------------
-            # work on a copy:
-            # so we can revert back to original `df` if something is failing
-            _df = df.copy()
-            # -----------------------------------------------------------------
-            # create a multi-index columns such as pint-pandas can work
-            _df.columns = pd.MultiIndex.from_tuples(
-                zip(_df.columns, _df.iloc[0].fillna(""))
+            # recover target units as specified in the validation schema
+            _props = schema["properties"]
+            self._target_units = {k: v for k in _props if (v := _props[k].get("units"))}
+            # split dataframe in two parts: one that will be quantified, the other not
+            # will be quantified the columns:
+            #  * have a specified unit (source units)
+            #  * are pd.to_numeric compliant
+            df_num, df_nonnum, self._source_units, u2cols = quantify_df(
+                df, self._target_units, early_report
             )
-            _df = _df.iloc[1:]
-            # -----------------------------------------------------------------
-            # collect defined units for later usage
-            self._source_units = dict(
-                _df.columns.tolist()
-            )  #  {'id': '', 'distA': 'm',... }
-            for i in range(len(_df.columns)):
-                try:
-                    _df.iloc[:, i] = pd.to_numeric(_df.iloc[:, i])
-                except:
-                    pass  # probably a column with strings
             try:
-                # insert the correct to workaround this bug:
-                # https://github.com/hgrecco/pint-pandas/issues/49
-                _df = _df.pint.quantify(level=-1)
-            except pint.errors.UndefinedUnitError as exc:
-                self._is_units = False
-                colnames = [
-                    colname
-                    for colname, unit in self._source_units.items()
-                    if unit in exc.args
-                ]
-                for colname in colnames:
-                    intermediate_report[(colname, "*")].append(
-                        "undefined units '%s'" % self._source_units[colname]
-                    )
-                    df = df[1:]
+                df_num = df_num.pint.quantify(level=-1)
             except Exception as exc:
+                early_report["uncatched unit error"].extend(list(exc.args))
                 self._is_units = False
-                intermediate_report["uncatched unit error"].extend(list(exc.args))
-            else:
-                # -------------------------------------------------------------
-                # everything was fine, commit temporary `_df` to `df`:
-                df = _df
-                del _df
+                df = df[1:]
         # =====================================================================
-        # first: fill empty values as requested by schema
+        # convert read units to schema expected units (if required)
+        # =====================================================================
+        if self._is_units:
+            # at this point, we still have df_num and df_nonnum
+            for col, units in self._target_units.copy().items():
+                if col not in df_num.columns:
+                    continue
+                df_num[col] = df_num[col].pint.to(units)
+            df = pd.concat((df_num.pint.dequantify(), df_nonnum), axis=1)
+            # re-arrange columns: restrict to level 0 and reorder as initially
+            df.columns = [t[0] for t in df.columns.tolist()]
+            df = df[[c for c in initial_without_urow.columns]]
+        # =====================================================================
+        # fill empty values as requested by schema
         # =====================================================================
         _fillnas = {
             k: schema["properties"][k]["items"].get("default")
@@ -385,41 +444,57 @@ class DataFrameSchematizer:
         if fillnas:
             df = df.fillna(value=fillnas, downcast="infer")
         # =====================================================================
-        # convert read units to schema expected units (if required)
+        # second validation
         # =====================================================================
-        if self._is_units:
-            self._target_units = {
-                k: schema["properties"][k].get("units", "")
-                for k in schema["properties"]
-            }
-            for col, units in self._target_units.items():
-                try:
-                    df[col] = df[col].pint.to(units)
-                except KeyError:
-                    pass
-                except AttributeError:
-                    breakpoint()
-            df = df.pint.dequantify()
-            df.columns = [t[0] for t in df.columns.tolist()]  # restrict to level 0
-
-        # =====================================================================
-        # second: validate
-        # =====================================================================
-        validator = jsonschema.Draft7Validator(schema)
         # df -> dict -> json -> dict to convert NaN to None
-        document = json.loads(json.dumps(df.to_dict(orient="list"), ignore_nan=True))
+        report = _validate(
+            json.loads(json.dumps(df.to_dict(orient="list"), ignore_nan=True)),
+            validator=validator,
+            check_types=False,
+            # report=early_report,
+        )
+        report = {**early_report, **report}
+        return df, len(report) == 0, dict(report)
+
+
+def _is_typing_error(error):
+    """check if error is type-checking error
+    return True if error is type-checking related
+    """
+    msgs = [error.message] + [e.message for e in error.context]
+    for msg in msgs:
+        if "is not of type" in msg:
+            return True
+    return False
+
+
+def _validate(document, validator, check_types=None, report=None):
+    """if check_types is:
+    * None (default) report everythong
+    * True: will only report typing
+    * False: will NOT report typing
+    """
+    if report is None:
         report = defaultdict(list)
-        for error in validator.iter_errors(document):
-            try:
-                col, row, *rows = error.absolute_path
-            except ValueError:
-                report["general"].append(error.message)
-            else:
+    for error in validator.iter_errors(document):
+        if _is_typing_error(error):
+            # if error reports a type checking issue, we skip in case
+            # `check_types` is False
+            if check_types is False:
+                continue
+        elif check_types is True:
+            # otherwise, for non-typing related issue, we skip if we **only**
+            # want typing-related issues
+            continue
+        try:
+            col, row, *rows = error.absolute_path
+        except ValueError:
+            report["general"].append(error.message)
+        else:
+            if error.message not in report[(col, row)]:
                 report[(col, row)].append(error.message)
                 report[(col, row)].extend([e.message for e in error.context])
-        report = {**dict(report), **intermediate_report}
-
-        return df, len(report) == 0, report
+    return report
 
 
 class ValidatorMixin:
@@ -429,12 +504,49 @@ class ValidatorMixin:
         """called by Base class __init__()"""
         self._schemas = {}
 
+    def getunits(self, tabname, key=None):
+        """retrieve from schema the working units. Raise KeyError the the
+        column is not units-aware.
+
+        if `key` is provided, the relevant units is returned (eg. "length" -> "mm")
+        otherwise, the whole dict is returned
+        """
+        props = self.get(tabname, "schemas").build()["properties"]
+        units = {
+            dim: units for dim, data in props.items() if (units := data.get("units"))
+        }
+        if key:
+            return units[key]
+        return units
+
+    def getrow(self, tabname, key, search_by=None):
+        """retrieve a row from any tabname, eventually sorting by `search_by`"""
+        df = self.get(tabname).copy()
+        if search_by:
+            df.set_index(search_by, inplace=True)
+        return df.loc[key]
+
+    def getvalue(self, tabname, key, column, search_by=None, as_quantity=True):
+        """retrive a cell content by intersecting row/column index
+
+        if as_quantity is:
+          * `True`: return a quantity argument *if possible*
+          * `False`: return magnitude
+        """
+        row = self.getrow(tabname, key, search_by)
+        value = row[column]
+        if as_quantity:
+            units = self.getunits(tabname, column)
+            return value * ureg.parse_expression(units)
+        else:
+            return value
+
     def quantify(self, tabname, restrict_to_units=False):
         df = self._data[tabname].copy()
         schema = self._schemas[tabname]
         if not schema._is_units:
             raise ValueError(f"tab {tabname} is not units-aware'")
-        units = schema._target_units
+        units = {k: v for k, v in schema._target_units.items() if v}
         # add dummy rows
         dummy_row = pd.DataFrame({c: [units.get(c, "")] for c in df.columns})
         df = pd.concat((dummy_row, df))
@@ -521,8 +633,13 @@ class ValidatorMixin:
         if not hasattr(self, "_raw_data"):
             self._raw_data = {tabname: df.copy() for tabname, df in self._data.items()}
         ret = {}
-        for tabname, df in self._data.items():
+        for tabname, df in self._data.copy().items():
             df, is_ok, report = self._validate_tab(tabname)
+            if df is None:
+                # tabname not described in schema; drop it
+                logging.info('drop tab "%s" which is not described in schema.', tabname)
+                self._data.pop(tabname)
+                continue
             self._data[tabname] = df  # override with filled (fillna) dataframe
             if not is_ok:
                 ret[tabname] = report
